//main.c
//Copyright (C) 2008 by raven
//This is Free Software released under the GNU/GPL License.

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <gccore.h>
#include <ogc/es.h>
#include <wiiuse/wpad.h>

#include "menu_tik_bin.h"
#include "menu_tmd_bin.h"
#include "menu_certs_bin.h"

#include "patch.h"
#include "patches.h"

typedef struct _dolheader
{
	u32 text_pos[7];
	u32 data_pos[11];
	u32 text_start[7];
	u32 data_start[11];
	u32 text_size[7];
	u32 data_size[11];
	u32 bss_start;
	u32 bss_size;
	u32 entry_point;
} dolheader;

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

void sort(u32 *a, int cnt);
void loadMenu();
void loadDol(u16 index);

void _unstub_start();

int main(int argc, char **argv)
{
	int i, idx, p[PATCH_COUNT];
	u64 *buf;
	s32 res, cios, ios, run, pad, pressed;
	u32 *ioses = NULL, tcnt = 0, icnt = 0;
	
	//Init stuff.
	VIDEO_Init();
	WPAD_Init();
	
	//Video and console setup.
	rmode = VIDEO_GetPreferredMode(NULL);
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
	VIDEO_Configure(rmode);
	VIDEO_SetNextFramebuffer(xfb);
	VIDEO_SetBlack(FALSE);
	VIDEO_Flush();
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE)
		VIDEO_WaitVSync();
	
	//Get stored IOS versions.
	res = ES_GetNumTitles(&tcnt);
	if(res < 0)
	{
		printf("ES_GetNumTitles: Error! (result = %d)\n", res);
		exit(0);
	}
	buf = memalign(32, sizeof(u64) * tcnt);
	res = ES_GetTitles(buf, tcnt);
	if(res < 0)
	{
		printf("ES_GetTitles: Error! (result = %d)\n", res);
		exit(0);
	}
	//Ugly...
	for(i = 0; i < tcnt; i++)
	{
		if(*((u32 *)(&(buf[i]))) == 1 && (u32)buf[i] > 2 && (u32)buf[i] < 0x100)
		{
			icnt++;
			ioses = (u32 *)realloc(ioses, sizeof(u32) * icnt);
			ioses[icnt - 1] = (u32)buf[i];
		}
	}
	free(buf);
	sort(ioses, icnt);
	
	printf("\n");
	printf("Menu loader (c) 2008 by raven\n");
	printf("Press [A] to launch the menu, [B] to exit.\n");
	
	memset(p, 0, sizeof(int) * PATCH_COUNT);
	cios = 0;
	idx = 0;
	run = 1;
	while(run == 1)
	{
		printf("\x1b[4;0H");
		
		if(idx == 0)
			printf(">> ");
		else
			printf("   ");
		printf("IOS Version: ");
		if(idx == 0)
			printf("\x1b[30m\x1b[47m IOS%-2d \x1b[37m\x1b[40m\n", ioses[cios]);
		else
			printf(" IOS%-2d \n", ioses[cios]);
		
		for(i = 0; i < PATCH_COUNT; i++)
		{
			if((i + 1) == idx)
				printf(">> ");
			else
				printf("   ");
			printf("%s: ", ptch[i].desc);
			if((i + 1) == idx)
				printf("\x1b[30m\x1b[47m %s \x1b[37m\x1b[40m\n", p[i] == 1 ? "ON " : "OFF");
			else
				printf(" %s \n", p[i] == 1 ? "ON " : "OFF");
		}
		
		pad = 0;
		while(pad == 0)
		{
			WPAD_ScanPads();
			pressed = WPAD_ButtonsDown(0);
			if(pressed)
				pad = 1;
			VIDEO_WaitVSync();
		}
		
		//Change option.
		if(pressed & WPAD_BUTTON_LEFT)
		{
			if(idx == 0)
			{
				if(cios > 0)
					cios--;
				else
					cios = icnt - 1;
			}
			else
			{
				if(p[idx - 1] == 0)
					p[idx - 1] = 1;
				else
					p[idx - 1] = 0;
			}
		}
		
		//Change option.
		if(pressed & WPAD_BUTTON_RIGHT)
		{
			if(idx == 0)
			{
				if(cios < icnt - 1)
					cios++;
				else
					cios = 0;
			}
			else
			{
				if(p[idx - 1] == 0)
					p[idx - 1] = 1;
				else
					p[idx - 1] = 0;
			}
		}
		
		//Down.
		if(pressed & WPAD_BUTTON_DOWN)
		{
			if(idx < PATCH_COUNT)
				idx++;
			else
				idx = 0;
		}
		
		//Up.
		if(pressed & WPAD_BUTTON_UP)
		{
			if(idx > 0)
				idx--;
			else
				idx = PATCH_COUNT;
		}
		
		//Launch.
		if(pressed & WPAD_BUTTON_A)
			run = 0;
		
		//Exit.
		if(pressed & WPAD_BUTTON_B)
		{
			printf("\nExiting...");
			exit(0);
		}
	}
	
	//Get the choosen IOS.
	ios = ioses[cios];
	free(ioses);
	
	WPAD_Shutdown();
	
	printf("\n");
	
	//Reload IOS.
	printf("Reloading IOS... ");
	res = IOS_ReloadIOS(ios);
	if(res < 0)
	{
		printf("Error! (result = %d)\n", res);
		exit(0);
	}
	printf("Done!\n");
	printf("IOS Version: IOS%2d %d.%d\n\n", 
			IOS_GetVersion(), 
			IOS_GetRevisionMajor(), 
			IOS_GetRevisionMinor());
	
	//Load menu.
	loadMenu();
	printf("\n");
	
	//Do patches.
	u32 cnt;
	for(i = 0; i < PATCH_COUNT; i++)
	{
		if(p[i] == 1)
		{
			printf("Patching '%s'... ", ptch[i].desc);
			cnt = do_patch((u8 *)0x8132ff80, 0x380000, &ptch[i]);
			printf("Done! (%d times)\n", cnt);
		}
	}
	printf("\n");
	
	printf("Launching the menu...\n");
	
	//Launch.
	SYS_ResetSystem(SYS_SHUTDOWN, 0, 0);
	DCFlushRange((u32*)0x8132ffe0, 0x380000);
	_unstub_start();
	
	printf("WTF?");
	while(1);
	
	return 0;
}

void sort(u32 *a, int cnt)
{
	u32 tmp;
	int i, j, m;
	
	for(i = 0; i < cnt; i++)
	{
		m = i;
		for(j = i; j < cnt; j++)
			if(a[j] > a[m])
				m = j;
		tmp = a[i];
		a[i] = a[m];
		a[m] = tmp;
	}
}

void loadMenu()
{
	u64 titleID;
	u32 tmdSize;
	u16 boot;
	signed_blob *ptmd;
	int res;
	
	//Identify as menu.
	printf("Identifying (1/2)... ");
	res = ES_Identify((signed_blob*)menu_certs_bin, menu_certs_bin_size, (signed_blob*)menu_tmd_bin, menu_tmd_bin_size, (signed_blob*)menu_tik_bin, menu_tik_bin_size, NULL);
	if(res < 0)
	{
		printf("Failed! (result = %d)\n", res);
		exit(0);
	}
	printf("Done!\n");
	
	//Get current title ID.
	ES_GetTitleID(&titleID);
	if(titleID != 0x0000000100000002ULL)
	{
		printf("Error! (TitleID = %016llx)\n", titleID);
		exit(0);
	}
	
	//Get tmd size.
	res = ES_GetStoredTMDSize(titleID, &tmdSize);
	if(res < 0)
	{
		printf("ES_GetStoredTMDSize: Error! (result = %d)\n", res);
		exit(0);
	}
	
	//Get tmd.
	ptmd = memalign(32, tmdSize);
	res = ES_GetStoredTMD(titleID, ptmd, tmdSize);
	if(res < 0)
	{
		printf("ES_GetStoredTMD: Error! (result = %d)\n", res);
		exit(0);
	}
	
	//Identify as menu.
	printf("Identifying (2/2)... ");
	res = ES_Identify((signed_blob*)menu_certs_bin, menu_certs_bin_size, ptmd, tmdSize, (signed_blob*)menu_tik_bin, menu_tik_bin_size, NULL);
	if(res < 0)
	{
		printf("Failed! (result = %d)\n", res);
		exit(0);
	}
	printf("Done!\n");
	
	ES_GetTitleID(&titleID);
	printf("TitleID: %016llx\n", titleID);
	
	printf("Version: %d\n", ((tmd *)SIGNATURE_PAYLOAD(ptmd))->title_version);
	
	boot = ((tmd *)SIGNATURE_PAYLOAD(ptmd))->boot_index;
	printf("Boot index: %d\n", boot);
	
	//int i;
	//u64 ds = (u64)-1;
	//for(i = 0; i < ((tmd *)SIGNATURE_PAYLOAD(ptmd))->num_contents; i++)
	//	if(TMD_CONTENTS(SIGNATURE_PAYLOAD(ptmd))[i].index == boot)
	//		ds = TMD_CONTENTS(SIGNATURE_PAYLOAD(ptmd))[i].size;
	//printf("Size: %lld\n", ds);
	
	printf("\n");
	
	loadDol(boot);
}

void loadDol(u16 index)
{
	u32 i;
	s32 s;
	//static dolheader _dolfile __attribute__((aligned(32)));
	//dolheader *dolfile = &_dolfile;
	dolheader *dolfile = (dolheader*)memalign(32, sizeof(dolheader));
	
	s = ES_OpenContent(index);
	if(s < 0)
	{
		printf("Error opening content! (result = %d)\n", s);
		return;
	}
	
	ES_ReadContent(s, (void *)dolfile, sizeof(dolheader));
	
	printf("Zeroing BSS at 0x%08x (0x%08x bytes).\n",
			dolfile->bss_start,
			dolfile->bss_size);
	
	memset((void *)dolfile->bss_start, 0, dolfile->bss_size);
	
	for (i = 0; i < 7; i++)
	{
		if(dolfile->data_start[i] < sizeof(dolheader))
			continue;
		
		printf("Loading TEXT section %u at 0x%08x (0x%08x bytes).\n",
				i, 
				dolfile->text_start[i],
				dolfile->text_size[i]);
		
		ES_SeekContent(s, dolfile->text_pos[i], 0);
		ES_ReadContent(s, (void *)dolfile->text_start[i], dolfile->text_size[i]);
	}
	
	for(i = 0; i < 11; i++)
	{
		if(dolfile->data_start[i] < sizeof(dolheader))
			continue;
		
		printf("Loading DATA section %u at 0x%08x (0x%08x bytes).\n",
				i, 
				dolfile->data_start[i],
				dolfile->data_size[i]);
		
		ES_SeekContent(s, dolfile->data_pos[i], 0);
		ES_ReadContent(s, (void *)dolfile->data_start[i], dolfile->data_size[i]);
	}
	
	ES_CloseContent(s);	
}
